@@ -1,19 +1,29 @@
 # coding=utf-8
 
+import functools
 import json
+import operator
 import pathlib
 import subprocess
 import sys
+import time
+import traceback
 from io import StringIO
 from os import environ, getenv, path
 from uuid import uuid4
 
+import colorama
 import pkg_resources
 import requests
 import yaml
 
 from kubedev.utils import KubedevConfig, YamlMerger
 
+colorama.init(autoreset=True)
+
+class RealSleep:
+  def sleep(self, seconds: float):
+    return time.sleep(seconds)
 
 class RealDownloader:
   def download_file_to(self, url: str, headers: dict, target_filename: str, file_accessor) -> bool:
@@ -43,9 +53,8 @@ class RealFileAccessor:
   def mkdirhier(self, path):
     return pathlib.Path(path).mkdir(parents=True, exist_ok=True)
 
-
 class RealShellExecutor:
-  def execute(self, commandWithArgs: list, envVars: dict = dict(), piped_input: str = None):
+  def execute(self, commandWithArgs: list, envVars: dict = dict(), piped_input: str = None, check=True):
     """
     Execute a shell command.
 
@@ -55,18 +64,20 @@ class RealShellExecutor:
     """
     cmds = [cmd for cmd in commandWithArgs if cmd is not None]
     print(
-      f'➡️   Executing "{" ".join(cmds)}" '
-      f'(additional env vars: {" ".join(envVars.keys())})',
+      f'{colorama.Fore.CYAN}➡️   Executing "{" ".join(cmds)}"' +
+      (f' (additional env vars: {" ".join(envVars.keys())})' if len(envVars) > 0 else ''),
       file=sys.stderr)
     return subprocess.run(cmds,
                           env      = {**environ, **envVars},
                           input    = piped_input if piped_input else None,
                           encoding = "UTF-8"     if piped_input else None,
+                          check    = check
                           ).returncode
 
-  def get_output(self, commandWithArgs):
+  def get_output(self, commandWithArgs, envVars: dict = dict(), check=True):
     cmds = [cmd for cmd in commandWithArgs if cmd is not None]
-    cmdResult = subprocess.run(cmds, check=True, stdout=subprocess.PIPE, encoding='utf-8')
+    print(f'{colorama.Fore.CYAN}➡️   Executing "{" ".join(cmds)}"')
+    cmdResult = subprocess.run(cmds, check=check, env=envVars, stdout=subprocess.PIPE, encoding='utf-8')
     if cmdResult.returncode == 0:
       return cmdResult.stdout
     else:
@@ -353,34 +364,36 @@ class Kubedev:
   def _template(self, kubedev, shell_executor, env_accessor, file_accessor, get_output=False):
     variables = KubedevConfig.get_global_variables(kubedev)
     tag = KubedevConfig.get_tag(env_accessor)
+    envs = KubedevConfig.get_helm_set_env_args(kubedev, env_accessor)
     command = [
         '/bin/sh',
         '-c',
         f'helm template ./helm-chart/ ' +
         f'--set KUBEDEV_TAG="{tag}"' +
-        KubedevConfig.get_helm_set_env_args(kubedev)
+        envs['cmdline']
     ]
     if not get_output:
-      return shell_executor.execute(command, variables)
+      return shell_executor.execute(command, {**variables, **envs['envs']})
     else:
-      return shell_executor.get_output(command)
+      return shell_executor.get_output(command, {**variables, **envs['envs']})
 
   def _deploy(self, kubedev, release_name, shell_executor, env_accessor, file_accessor, get_output=False):
     variables = KubedevConfig.get_global_variables(kubedev)
     tag = KubedevConfig.get_tag(env_accessor)
     kubeconfig = KubedevConfig.get_kubeconfig_path(env_accessor, file_accessor)
+    envs = KubedevConfig.get_helm_set_env_args(kubedev, env_accessor)
     command = [
         '/bin/sh',
         '-c',
         f'helm upgrade {release_name} ./helm-chart/ --install --wait ' +
         f'--kubeconfig {kubeconfig} {self._get_kubecontext_arg(env_accessor)} ' +
         f'--set KUBEDEV_TAG="{tag}"' +
-        KubedevConfig.get_helm_set_env_args(kubedev)
+        envs['cmdline']
     ]
     if not get_output:
-      return shell_executor.execute(command, variables)
+      return shell_executor.execute(command, {**variables, **envs['envs']})
     else:
-      return shell_executor.get_output(command)
+      return shell_executor.get_output(command, {**variables, **envs['envs']})
 
   def template(self, configFileName, shell_executor=RealShellExecutor(), env_accessor=RealEnvAccessor(), file_accessor=RealFileAccessor()):
     return self.template_from_config(
@@ -413,11 +426,8 @@ class Kubedev:
     if envCi is not None and envDockerAuthConfig is not None and envHome is not None:
       dockerConfigPath = path.join(envHome, '.docker/config.json')
       if file_accessor.load_file(dockerConfigPath) is None:
-        print('NOTICE NOTICE NOTICE')
-        print('CI environment detected and no docker config found.')
-        print(f'Storing content of ${{DOCKER_AUTH_CONFIG}} to file {dockerConfigPath}.')
-        print('NOTICE NOTICE NOTICE')
-        print()
+        print(f'{colorama.Fore.YELLOW}CI environment detected and no docker config found.')
+        print(f'{colorama.Fore.YELLOW}Storing content of ${{DOCKER_AUTH_CONFIG}} to file {dockerConfigPath}.')
         file_accessor.save_file(dockerConfigPath, envDockerAuthConfig, overwrite=False)
         return True
     return False
@@ -440,14 +450,15 @@ class Kubedev:
         imageTag = image['imageName']
       else:
         imageTag = f"{image['imageNameTagless']}:{force_tag}"
+      (argsCmdLine, extraEnv) = KubedevConfig.get_docker_build_args(image, env_accessor=env_accessor)
       call = [
           '/bin/sh',
           '-c',
           f"docker build -t {imageTag} " +
-          KubedevConfig.get_docker_build_args(image) +
+          argsCmdLine +
           f"{image['buildPath']}"
       ]
-      return shell_executor.execute(call, dict())
+      return shell_executor.execute(call, envVars=extraEnv, check=False)
 
   def push(self, configFileName, container, file_accessor=RealFileAccessor(), shell_executor=RealShellExecutor(), env_accessor=RealEnvAccessor()):
     return self.push_from_config(
@@ -572,8 +583,8 @@ class Kubedev:
             '❌ Required field "imagePullSecrets" is missing in kubedev.json', True)
         result = False
 
-    envs = KubedevConfig.get_all_env_names(kubedev, build=is_command(
-        'build'), container=is_command('deploy') or is_command('template'))
+    envs = KubedevConfig.get_all_envs(kubedev, build=is_command(
+        'build'), container=is_command('deploy') or is_command('template')).keys()
     for env in sorted(envs):
       if isinstance(env_accessor.getenv(env), type(None)):
         printer.print(
@@ -613,14 +624,218 @@ class Kubedev:
       if buildResult != 0:
         return buildResult
       else:
+        (runEnvArgs, extraEnvs) = KubedevConfig.get_docker_run_envs(image, env_accessor=env_accessor)
+
         command = [
           '/bin/sh',
           '-c',
           f"docker run --interactive {interactive_flags}--rm " +
           KubedevConfig.get_docker_run_volumes(image, file_accessor, shell_executor) +
           KubedevConfig.get_docker_run_ports(image) +
-          KubedevConfig.get_docker_run_envs(image) +
+          runEnvArgs +
           f"{image['imageNameTagless']}:{currentTag}"
         ]
-        runResult = shell_executor.execute(command, dict())
-        return runResult
+        return shell_executor.execute(command, envVars=extraEnvs, check=False)
+
+  @staticmethod
+  def _run_docker_detached(network: str, name: str, ports: list, rawImage: str, images: dict, variables: dict, shell_executor: object) -> (str, bool):
+      """
+      Starts a container in detached mode and returns it's ID.
+
+      Returns None when the start failed.
+      """
+      (image, requiredEnvs, isFromKubedev) = Kubedev._build_image(rawImage, images)
+      filteredRequiredEnvs = sorted([env for env in requiredEnvs if not env in variables]) # Note: the sorted() is important, otherwise the order would be
+      print(f'Running detached: {name} (image: {image})')
+      cmdRm = ["docker", "rm", "--force", name]
+      shell_executor.execute(cmdRm, check=False) # To be sure, first try to delete the container that we want to create
+      requiredEnvForwards = functools.reduce(operator.concat, [["--env", f'{envName}="${{{envName}}}"'] for envName in filteredRequiredEnvs], []) if isFromKubedev else []
+      cmdCreate = [
+        "/bin/sh",
+        "-c",
+        " ".join([
+          "docker",
+          "create",
+          "--network", network,
+          "--name", name,
+          "--rm"] + \
+          requiredEnvForwards + \
+          functools.reduce(operator.concat, [["--env", f'{varName}="{varValue}"'] for varName, varValue in variables.items()], []) + \
+          functools.reduce(operator.concat, [["--publish", str(port)] for port in ports], []) + \
+          [image])]
+      dockerIdRaw = shell_executor.get_output(cmdCreate, check=False)
+      print(f"> {dockerIdRaw}")
+      if dockerIdRaw is None or dockerIdRaw == "":
+          return (dockerIdRaw, name, False)
+      else:
+          dockerId = dockerIdRaw.strip(" \r\n\t")
+          cmdStart = ["docker", "start", dockerId]
+          if shell_executor.execute(cmdStart, check=False) == 0:
+              return (dockerId, name, True)
+          else:
+              return (dockerId, name, False)
+
+  @staticmethod
+  def _build_image(name: str, images: dict) -> (str, bool):
+      if len(name) > 3 and name[0] == '{' and name[-1] == '}':
+          appName = name[1:-1]
+          if appName in images:
+              return (images[appName]['imageName'], images[appName]['containerEnvs'].keys(), True)
+          else:
+              raise Exception(f'App "{appName}" is referenced by the system test service {name}, but is not defined in kubedev config')
+      return (name, dict(), False)
+
+  @staticmethod
+  def _field_required(obj: dict, field: str, objectName: str):
+      if not field in obj:
+          raise Exception(f'The field {field} is required in {objectName}')
+      else:
+          return obj[field]
+
+  @staticmethod
+  def _field_optional(obj: dict, field: str, default):
+      if not field in obj:
+          return default
+      else:
+          return obj[field]
+
+  def system_test(self,
+                  configFileName: str,
+                  appName: str,
+                  file_accessor=RealFileAccessor(),
+                  env_accessor=RealEnvAccessor(),
+                  shell_executor=RealShellExecutor(),
+                  tag_generator=TagGenerator(),
+                  sleeper=RealSleep()) -> bool:
+    return self.system_test_from_config(
+      self._load_config(configFileName, file_accessor),
+      appName,
+      file_accessor=file_accessor,
+      env_accessor=env_accessor,
+      shell_executor=shell_executor,
+      tag_generator=tag_generator,
+      sleeper=sleeper)
+
+  def system_test_from_config(self, kubedev, appName: str, file_accessor, env_accessor, shell_executor, tag_generator, sleeper) -> bool:
+      '''
+      Runs the system tests for an app as defined in the kubedev config.
+
+      @param requiredEnvs Are the required-envs from the global and the deployment level that this system test definition comes from
+      @param images Is the list of all available images from apps defined in the kubedev config
+      @param systemTestDefinition Is the node "systemTests" from this apps definition
+      @param shell_executor Is used to execute shell commands
+      '''
+
+      self._create_docker_config(file_accessor, env_accessor)
+
+      images = KubedevConfig.get_images(kubedev, env_accessor)
+      apps = KubedevConfig.get_all_app_definitions(kubedev)
+      if not appName in apps:
+        print(f'{colorama.Fore.RED} Invalid app {appName} specified. Available apps: {apps.keys()}', file=sys.stderr)
+        return False
+
+      app = apps[appName]
+      if not 'systemTest' in app:
+        print(f'{colorama.Fore.RED} App {appName} does not define a systemTest.', file=sys.stderr)
+
+      systemTestDefinition = app['systemTest']
+
+      # Step #1: Build the system test container
+      globalVariables = self._field_optional(systemTestDefinition, "variables", dict())
+
+      testContainer = self._field_optional(systemTestDefinition, "testContainer", dict())
+      buildArgs = self._field_optional(testContainer, "buildArgs", dict())
+      variables = {**globalVariables, **self._field_optional(testContainer, "variables", dict())}
+      requiredEnvs = images[appName]['containerEnvs']
+      (filteredRequiredEnvs, additionalEnvs) = KubedevConfig.prepare_envs(requiredEnvs, env_accessor)
+
+      containerDir = f"./systemTests/{appName}/"
+      uuid = tag_generator.tag()
+      tag = f"local-{appName}-system-tests-{uuid}"
+      network = tag
+      cmdBuild = [
+        "/bin/sh",
+        "-c",
+        " ".join([
+          "docker",
+          "build",
+          "-t",
+          tag
+          ] + functools.reduce(operator.concat, [["--build-args", f'{arg}="{value}"'] for arg, value in buildArgs.items()], []) + [containerDir])]
+      if shell_executor.execute(cmdBuild, envVars=buildArgs, check=False) != 0:
+          return False
+
+      # Step #2: Create the docker network
+      cmdNetworkCreate = ["docker", "network", "create", network]
+      if shell_executor.execute(cmdNetworkCreate, check=False) != 0:
+          return False
+
+      result = False
+      startedContainers = []
+      try:
+          # Step #3: Start the service container
+          startedContainers = [
+              self._run_docker_detached(
+                  network,
+                  KubedevConfig.expand_variables(self._field_required(service, 'hostname', 'systemTest.service'), env_accessor, self._field_optional(service, 'variables', dict())),
+                  self._field_required(service, 'ports', 'systemTest.service'),
+                  serviceKey,
+                  images,
+                  {**globalVariables, **self._field_optional(service, 'variables', dict())},
+                  shell_executor) for serviceKey, service in self._field_optional(systemTestDefinition, 'services', dict()).items()]
+
+          numSleepSeconds = 5
+          print(f'{colorama.Fore.YELLOW}TODO: Sleeping for {numSleepSeconds} seconds instead of pinging the exposed ports')
+          # Step #4: Wait for the services to become ready
+          sleeper.sleep(numSleepSeconds)
+
+          # Step #5: Run the system test container
+          cmdRunSystemTests = [
+              "/bin/sh",
+              "-c",
+              " ".join([
+              "docker",
+              "run",
+              "--rm",
+              "--network", network,
+              "--name", f"{appName}-system-tests-{uuid}",
+              "--interactive"] + \
+              functools.reduce(operator.concat, [["--env", f'{envName}="${{{attribs["targetName"]}}}"'] for envName, attribs in filteredRequiredEnvs.items()], []) + \
+              functools.reduce(operator.concat, [["--env", f'{varName}="{varValue}"'] for varName, varValue in variables.items()], []) + \
+              [tag])]
+          if shell_executor.execute(cmdRunSystemTests, envVars=additionalEnvs, check=False) == 0:
+              result = True
+          else:
+              print()
+              print(f'{colorama.Fore.RED}^^^ See logs of the system test above.')
+              print()
+              result = False
+          return result
+      except:
+          traceback.print_exc()
+      finally:
+          for startedContainer in startedContainers:
+              containerId = startedContainer[0]
+              if containerId != None and containerId != "":
+                  if result == False:
+                      cmdLogs = ["docker", "logs", containerId]
+                      shell_executor.execute(cmdLogs, check=False)
+                      print()
+                      print(f'{colorama.Fore.RED}^^^ See logs of the service "{startedContainer[1]}" above')
+                      print()
+                  # Cleanup #2: Remove the service containers
+                  cmdRm = ["docker", "rm", "--force", containerId]
+                  shell_executor.execute(cmdRm, check=False)
+          # Cleanup #2: Remove the docker network
+          cmdNetworkRm = ["docker", "network", "rm", tag]
+          shell_executor.execute(cmdNetworkRm, check=False)
+          if result:
+              print()
+              print(f'{colorama.Fore.GREEN}System tests succeeded! 🌟🎉🥳')
+              print()
+              return 0
+          else:
+              print()
+              print(f'{colorama.Fore.RED}System tests failed! The logs of the services and the system test have been printed above.')
+              print()
+              return 1
